from pathlib import Path

import pytest

from hyprshade.cli.utils import ls_dirs, write_systemd_user_unit


class TestLsDirs:
    def test_empty(self):
        assert list(ls_dirs([])) == []

    def test_single(self, tmp_path: Path):
        (tmp_path / "foo").touch()
        assert list(ls_dirs([tmp_path])) == ["foo"]

    def test_multiple(self, tmp_path: Path):
        (tmp_path / "foo").touch()
        (tmp_path / "bar").touch()
        assert list(ls_dirs([tmp_path])) == ["bar", "foo"]

    def test_multiple_dirs(self, tmp_path_factory: pytest.TempPathFactory):
        paths = []
        for dir_ in ["foo", "bar", "baz"]:
            tmp_path_ = tmp_path_factory.mktemp(dir_)
            (tmp_path_ / f"qux{dir_}").touch()
            paths.append(tmp_path_)

        assert list(ls_dirs(paths)) == ["quxbar", "quxbaz", "quxfoo"]

    def test_unique(self, tmp_path_factory: pytest.TempPathFactory):
        paths = []
        for dir_ in ["foo", "bar", "baz"]:
            tmp_path_ = tmp_path_factory.mktemp(dir_)
            (tmp_path_ / "qux").touch()
            paths.append(tmp_path_)

        assert list(ls_dirs(paths)) == ["qux"]


def test_write_systemd_user_unit(monkeypatch: pytest.MonkeyPatch, tmp_path: Path):
    monkeypatch.setenv("XDG_CONFIG_HOME", str(tmp_path))
    (tmp_path / "systemd" / "user").mkdir(parents=True)

    write_systemd_user_unit("service", "foo")
    assert (tmp_path / "systemd" / "user" / "hyprshade.service").read_text() == "foo"

    write_systemd_user_unit("timer", "bar")
    assert (tmp_path / "systemd" / "user" / "hyprshade.timer").read_text() == "bar"



from pathlib import Path

import pytest
from click.testing import CliRunner

from hyprshade.cli import cli
from hyprshade.shader import hyprctl

pytestmark = [
    pytest.mark.requires_hyprland(),
    pytest.mark.usefixtures("_save_screen_shader"),
]


def test_off(runner: CliRunner):
    result = runner.invoke(cli, "off")

    assert result.exit_code == 0
    assert hyprctl.get_screen_shader() is None


def test_off_after_set(runner: CliRunner, shader_path_env: Path):
    hyprctl.set_screen_shader(shader_path_env.as_posix())
    result = runner.invoke(cli, "off")

    assert result.exit_code == 0
    assert hyprctl.get_screen_shader() is None



import os
import sysconfig
from collections.abc import Callable
from functools import lru_cache
from pathlib import Path

import pytest

from hyprshade.shader import Shader, hyprctl


@pytest.fixture(scope="module")
def _save_screen_shader():
    """Save the current screen shader and restore it after testing"""

    screen_shader = hyprctl.get_screen_shader()
    yield

    if screen_shader is None:
        hyprctl.clear_screen_shader()
        return
    try:
        hyprctl.set_screen_shader(screen_shader)
    except BaseException:
        import os

        hyprctl.clear_screen_shader()
        os.system('notify-send "hyprshade" "Failed to restore screen shader"')


@pytest.fixture()
def _clear_shader_env(shader_dir_env, shader_dir_user, shader_dir_system):
    pass


@pytest.fixture()
def _clear_screen_shader():
    """Clear the current screen shader before and after each test"""

    hyprctl.clear_screen_shader()
    yield
    hyprctl.clear_screen_shader()


@pytest.fixture()
def shader_dir_env(monkeypatch: pytest.MonkeyPatch, tmp_path: Path):
    prev_env = os.environ.get(Shader.dirs.ENV_VAR_NAME)

    path_ = tmp_path / "env/shaders"
    path_.mkdir(parents=True, exist_ok=True)
    monkeypatch.setenv(Shader.dirs.ENV_VAR_NAME, str(path_))
    yield path_

    if prev_env is None:
        monkeypatch.delenv(Shader.dirs.ENV_VAR_NAME, raising=False)
    else:
        monkeypatch.setenv(Shader.dirs.ENV_VAR_NAME, prev_env)


@pytest.fixture()
def shader_dir_user(monkeypatch: pytest.MonkeyPatch, tmp_path: Path):
    xdg_config_home = os.environ.get("XDG_CONFIG_HOME")

    config_path = tmp_path / "config"
    monkeypatch.setenv("XDG_CONFIG_HOME", str(config_path))

    path_ = config_path / "hypr/shaders"
    path_.mkdir(parents=True, exist_ok=True)
    yield path_

    if xdg_config_home is None:
        monkeypatch.delenv("XDG_CONFIG_HOME", raising=False)
    else:
        monkeypatch.setenv("XDG_CONFIG_HOME", xdg_config_home)


@pytest.fixture()
def shader_dir_system(monkeypatch: pytest.MonkeyPatch, tmp_path: Path):
    prev_sysconfig_get_path = sysconfig.get_path

    data_path = tmp_path / "data"
    monkeypatch.setattr(
        sysconfig, "get_path", lambda name: str(data_path) if name == "data" else ""
    )

    path_ = Path(sysconfig.get_path("data"), "share", "hyprshade", "shaders").resolve()
    path_.mkdir(parents=True, exist_ok=True)
    yield path_

    monkeypatch.setattr(sysconfig, "get_path", prev_sysconfig_get_path)


def _shader_path_at(dir_: Path) -> Path:
    path_ = dir_ / "shader.glsl"
    path_.write_text("void main() {}")
    return path_


@pytest.fixture()
def shader_path(tmp_path: Path) -> Path:
    return _shader_path_at(tmp_path)


@pytest.fixture()
def shader_path_env(shader_dir_env: Path) -> Path:
    return _shader_path_at(shader_dir_env)


@pytest.fixture()
def shader_path_user(shader_dir_user: Path) -> Path:
    return _shader_path_at(shader_dir_user)


@pytest.fixture()
def shader_path_system(shader_dir_system: Path) -> Path:
    return _shader_path_at(shader_dir_system)


@pytest.fixture()
def shader_path_factory(shader_dir_env: Path) -> Callable[[str], Path]:
    def _shader_path(name: str) -> Path:
        path_ = shader_dir_env / f"{name}.glsl"
        path_.write_text("void main() {}")
        return path_

    return _shader_path


def pytest_runtest_setup(item: pytest.Item) -> None:
    for marker in item.iter_markers():
        if (
            marker.name == "requires_hyprland"
            and not has_hyprland()
            and not item.config.getoption("--hyprland", default=False)
        ):
            pytest.skip("Not running in hyprland")


@lru_cache
def has_hyprland():
    return os.getenv("HYPRLAND_INSTANCE_SIGNATURE") is not None



# mypy: disable-error-code="arg-type"

from pathlib import Path

import click
import pytest

from hyprshade.cli import utils


class TestConvertToShader:
    def test_works(self):
        shader = utils.convert_to_shader(None, None, "foo")
        assert shader is not None
        assert shader.name == "foo"

    def test_none(self):
        assert utils.convert_to_shader(None, None, None) is None

    def test_stale(self):
        if Path("./foo").exists():
            pytest.fail("test assumption failed: ./foo exists")

        with pytest.raises(click.BadParameter):
            utils.convert_to_shader(None, None, "./foo")


class TestValidateOptionalParam:
    def test_empty(self):
        assert utils.validate_optional_param(None, None, ()) is None

    def test_single(self):
        assert utils.validate_optional_param(None, None, ("foo",)) == "foo"

    def test_rejects_multiple(self):
        with pytest.raises(click.UsageError):
            utils.validate_optional_param(None, None, ("foo", "bar"))


class TestOptionalParam:
    def test_basic(self):
        o = utils.optional_param("FOO", None)
        assert o["metavar"] == "FOO"
        assert o["nargs"] == -1
        assert o["callback"] is not None

    def test_callback(self):
        o = utils.optional_param("FOO", None)
        callback = o["callback"]
        assert callback(None, None, ("foo",)) == "foo"
        assert callback(None, None, ()) is None

    def test_given_callback(self):
        o = utils.optional_param("FOO", lambda *_: "bar")
        callback = o["callback"]
        assert callback(None, None, ("foo",)) == "bar"

    def test_callback_synergy(self):
        o = utils.optional_param("FOO", lambda _1, _2, x: x + "bar")
        callback = o["callback"]
        assert callback(None, None, ("foo",)) == "foobar"



from __future__ import annotations

from itertools import chain, pairwise
from typing import TYPE_CHECKING

from hyprshade.shader import Shader
from hyprshade.utils import is_time_between

if TYPE_CHECKING:
    from collections.abc import Iterator
    from datetime import time

    from .core import Config
    from .types import ShaderConfig, TimeRange


class ScheduleEntry:
    shader: Shader
    start_time: time
    end_time: time | None

    def __init__(self, shader_config: ShaderConfig):
        self.shader = Shader(shader_config["name"])
        self.start_time = shader_config["start_time"]
        self.end_time = shader_config.get("end_time")

    def __repr__(self) -> str:
        return (
            f'ScheduleEntry("{self.shader}", start_time={self.start_time}, '
            f"end_time={self.end_time})"
        )


class Schedule:
    entries: list[ScheduleEntry]
    default_shader: Shader | None

    def __init__(self, config: Config):
        rest_shader_configs, default_shader_config = config.partition()
        sorted_rest_shader_configs = sorted(
            rest_shader_configs, key=lambda s: s["start_time"]
        )
        self.entries = list(map(ScheduleEntry, sorted_rest_shader_configs))
        self.default_shader = (
            Shader(default_shader_config["name"])
            if default_shader_config is not None
            else None
        )

    def scheduled_shader(self, t: time) -> Shader | None:
        for shader, (start_time, end_time) in self._resolved_entries():
            if is_time_between(t, start_time, end_time):
                return shader

        return self.default_shader

    def event_times(self) -> Iterator[time]:
        for entry in self.entries:
            yield entry.start_time
            if entry.end_time is not None:
                yield entry.end_time

    def _resolved_entries(self) -> Iterator[tuple[Shader, TimeRange]]:
        for entry, next_entry in pairwise(chain(self.entries, [self.entries[0]])):
            start_time = entry.start_time
            end_time = entry.end_time or next_entry.start_time
            yield entry.shader, (start_time, end_time)

    @staticmethod
    def from_config(path_: str | None = None) -> Schedule:
        from .core import Config

        return Schedule(Config(path_))



from collections.abc import Callable
from pathlib import Path

import pytest
from click.testing import CliRunner

from hyprshade.cli import cli
from hyprshade.shader import hyprctl

pytestmark = [
    pytest.mark.requires_hyprland(),
    pytest.mark.usefixtures("_save_screen_shader"),
    pytest.mark.usefixtures("_clear_shader_env"),
]


class TestCurrent:
    def test_empty(self, runner: CliRunner, shader_path_env: Path):
        hyprctl.clear_screen_shader()

        result = runner.invoke(cli, ["current"])

        assert result.exit_code == 0
        assert result.output == ""

    def test_it(self, runner: CliRunner, shader_path_factory: Callable[[str], Path]):
        shader_path_factory("shader1")

        hyprctl.set_screen_shader("shader1")

        result = runner.invoke(cli, ["current"])

        assert result.exit_code == 0
        assert result.output.strip() == "shader1"



from __future__ import annotations

import shlex

import click

from hyprshade.config.schedule import Schedule

from .utils import get_script_path, write_systemd_user_unit


@click.command(short_help="Install systemd user units")
def install():
    """Install systemd user units.

    Requires a schedule to be specified in hyprshade.toml.
    """

    script_path = get_script_path()
    schedule = Schedule.from_config()
    timer_config = "\n".join(
        sorted([f"OnCalendar=*-*-* {x}" for x in schedule.event_times()])
    )

    write_systemd_user_unit(
        "service",
        f"""[Unit]
Description=Apply screen filter

[Service]
Type=oneshot
ExecStart={shlex.quote(script_path)} auto
""",
    )

    write_systemd_user_unit(
        "timer",
        f"""[Unit]
Description=Apply screen filter on schedule

[Timer]
{timer_config}

[Install]
WantedBy=timers.target
""",
    )



from __future__ import annotations

import os
import sys
from os import PathLike, path
from typing import TYPE_CHECKING, Any, Literal, TypeVar

import click
from more_itertools import flatten, unique_justseen

from hyprshade.config.utils import systemd_user_config_home
from hyprshade.shader import Shader

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Iterator

T = TypeVar("T", str, int, float, bool, click.ParamType)
SystemdUnitType = Literal["service", "timer"]


def convert_to_shader(
    ctx: click.Context, param: click.Argument, shader_name_or_path: str | None
) -> Shader | None:
    shader = Shader(shader_name_or_path) if shader_name_or_path is not None else None
    if shader is not None and shader.stale:
        raise click.BadParameter(f"Shader {shader_name_or_path} does not exist")
    return shader


def validate_optional_param(
    ctx: click.Context, param: click.Argument, value: tuple[T, ...]
) -> T | None:
    if extra_args := value[1:]:
        s = "s" if len(extra_args) > 1 else ""
        raise click.UsageError(
            f"Got unexpected extra argument{s} ({' '.join(map(str, extra_args))})"
        )

    return None if len(value) == 0 else value[0]


def optional_param(
    metavar: str | None = None, callback: Callable | None = None
) -> dict[str, Any]:
    def merged_callback(
        ctx: click.Context, param: click.Argument, value: tuple[T, ...]
    ):
        value2 = validate_optional_param(ctx, param, value)
        if callback is not None:
            return callback(ctx, param, value2)
        return value2

    return {
        "metavar": metavar,
        "nargs": -1,
        "callback": merged_callback,
    }


def ls_dirs(dirs: Iterable[str | PathLike[str]]) -> Iterator[str]:
    return unique_justseen(sorted(flatten(map(os.listdir, dirs))))


def write_systemd_user_unit(unit_type: SystemdUnitType, body: str) -> None:
    dest_dir = systemd_user_config_home()
    os.makedirs(dest_dir, exist_ok=True)
    with open(path.join(dest_dir, f"hyprshade.{unit_type}"), "w") as f:
        f.write(body)


def get_script_path() -> str:
    return path.realpath(sys.argv[0], strict=True)



from __future__ import annotations

from datetime import datetime, time

import click
from more_itertools import quantify

from hyprshade.config.schedule import Schedule
from hyprshade.shader import Shader

from .utils import convert_to_shader, optional_param


def get_shader_to_toggle(
    shader: Shader | None, fallback: Shader | None
) -> Shader | None:
    current = Shader.current()
    if shader != current:
        return shader
    return fallback


def get_fallback(
    *,
    shader: Shader | None,
    default: Shader | None,
    auto: Shader | None,
    fallback_default: bool,
    fallback_auto: bool,
) -> Shader | None:
    if fallback_default or (fallback_auto and shader == auto):
        return default
    elif fallback_auto:
        return auto
    return None


def try_from_config(t: time, *, panic: bool) -> tuple[Shader | None, Shader | None]:
    try:
        schedule = Schedule.from_config()
    except FileNotFoundError:
        if panic:
            raise
        return None, None
    return schedule.scheduled_shader(t), schedule.default_shader


@click.command(short_help="Toggle screen shader")
@click.argument("shader", **optional_param("SHADER", convert_to_shader))
@click.option(
    "--fallback",
    metavar="SHADER",
    callback=convert_to_shader,
    help="Select fallback shader",
)
@click.option(
    "--fallback-default",
    is_flag=True,
    default=False,
    help="Use default shader as fallback",
)
@click.option(
    "--fallback-auto",
    is_flag=True,
    default=False,
    help="Automatically infer fallback",
)
def toggle(
    shader: Shader | None,
    fallback: Shader | None,
    fallback_default: bool,
    fallback_auto: bool,
):
    """Toggle screen shader.

    The default behavior is to toggle between SHADER and off. If SHADER is
    not provided, it is inferred from schedule configuration.

    When a fallback shader is provided with one of the --fallback* options, the
    toggle will be between SHADER and the fallback shader.

    --fallback-auto will determine the fallback from the schedule configuration.
    If the currently scheduled shader and SHADER are identical, the fallback
    will instead be the default shader.
    """

    t = datetime.now().time()

    fallback_opts = [fallback, fallback_default, fallback_auto]
    if quantify(fallback_opts) > 1:
        raise click.BadOptionUsage(
            "--fallback", "Must not specify more than one --fallback* option"
        )

    scheduled, default = try_from_config(t, panic=(fallback_default or fallback_auto))
    shader = shader or scheduled

    fallback = fallback or get_fallback(
        shader=shader,
        default=default,
        auto=scheduled,
        fallback_default=fallback_default,
        fallback_auto=fallback_auto,
    )
    shader_to_toggle = get_shader_to_toggle(shader, fallback)
    if shader_to_toggle:
        shader_to_toggle.on()
    else:
        Shader.off()



from __future__ import annotations

import click

from hyprshade.shader import Shader


@click.command(short_help="Print current screen shader")
@click.option("-l", "--long", is_flag=True, help="Long listing format")
def current(long: bool):
    """Print current screen shader.

    If no shader is active, print nothing.
    """

    current = Shader.current()
    if current is None:
        return

    if long:
        click.echo(f"{current}  {current.dirname()}")
        return
    click.echo(current)



from datetime import time

from hyprshade import utils


class TestIsTimeBetween:
    def test_normal(self):
        assert utils.is_time_between(time(12, 0), time(11, 0), time(13, 0)) is True

    def test_wrapping(self):
        assert utils.is_time_between(time(1, 0), time(20, 0), time(2, 0)) is True
        assert utils.is_time_between(time(23, 0), time(20, 0), time(2, 0)) is True



from __future__ import annotations

import logging
from os import path
from typing import Final

from more_itertools import first_true, flatten

from hyprshade.config.utils import hypr_config_home, hyprshade_config_home
from hyprshade.utils import scandir_recursive

from . import hyprctl


def _stripped_basename(s: str) -> str:
    return path.splitext(path.basename(s))[0]


class _ShaderDirs:
    ENV_VAR_NAME: Final = "HYPRSHADE_SHADERS_DIR"
    SYSTEM_DIR: Final = "/usr/share/hyprshade/shaders"

    @staticmethod
    def env() -> str:
        return path.expanduser(
            path.expandvars(path.expandvars("$" + _ShaderDirs.ENV_VAR_NAME))
        )

    @staticmethod
    def user_hypr() -> str:
        return path.join(hypr_config_home(), "shaders")

    @staticmethod
    def user_hyprshade() -> str:
        return path.join(hyprshade_config_home(), "shaders")

    @staticmethod
    def system() -> str:
        import sysconfig

        return first_true(
            [path.join(sysconfig.get_path("data"), "share", "hyprshade", "shaders")],
            pred=path.exists,
            default=_ShaderDirs.SYSTEM_DIR,
        )

    @staticmethod
    def all() -> list[str]:
        return [
            x
            for x in [
                _ShaderDirs.env(),
                _ShaderDirs.user_hypr(),
                _ShaderDirs.user_hyprshade(),
                _ShaderDirs.system(),
            ]
            if path.exists(x)
        ]


class Shader:
    dirs: Final = _ShaderDirs

    def __init__(self, shader_name_or_path: str):
        self._given_path = (
            path.abspath(shader_name_or_path)
            if shader_name_or_path.find(path.sep) != -1
            else None
        )
        self._name = _stripped_basename(shader_name_or_path)
        self._stale = self._given_path is not None and not path.exists(self._given_path)

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, Shader):
            return False
        try:
            s, s2 = self._resolve_path(), __value._resolve_path()
        except FileNotFoundError:
            return False
        return path.samefile(s, s2)

    def __str__(self) -> str:
        return self._name

    def __repr__(self) -> str:
        return f"Shader({self._name!r})"

    @property
    def name(self) -> str:
        return self._name

    @property
    def stale(self) -> bool:
        return self._stale

    def dirname(self) -> str:
        return path.dirname(self._resolve_path())

    def on(self) -> None:
        path_ = self._resolve_path()
        logging.debug(f"Turning on shader '{self._name}' at '{path_}'")
        hyprctl.set_screen_shader(path_)

    @staticmethod
    def off() -> None:
        hyprctl.clear_screen_shader()

    @staticmethod
    def current() -> Shader | None:
        path_ = hyprctl.get_screen_shader()
        return None if path_ is None else Shader(path_)

    def _resolve_path(self) -> str:
        if self._stale:
            raise FileNotFoundError(
                f"Shader '{self._name}' does not exist at '{self._given_path}'"
            )
        if self._given_path:
            return self._given_path

        dirs = Shader.dirs.all()
        all_files = flatten(scandir_recursive(d, max_depth=5) for d in dirs)
        for file in all_files:
            if path.splitext(file.name)[0] == self._name:
                return file.path

        raise FileNotFoundError(
            f"Shader '{self._name}' could not be found in any of the following"
            " directories:\n\t"
            "{}".format("\n\t".join(dirs))
        )



import re
from collections.abc import Callable
from pathlib import Path

import pytest
from click.testing import CliRunner

from hyprshade.cli import cli
from hyprshade.shader import hyprctl

pytestmark = [
    pytest.mark.requires_hyprland(),
    pytest.mark.usefixtures("_save_screen_shader"),
    pytest.mark.usefixtures("_clear_shader_env"),
    pytest.mark.usefixtures("_clear_screen_shader"),
]


class TestLs:
    def test_empty(self, runner: CliRunner):
        result = runner.invoke(cli, ["ls"])

        assert result.exit_code == 0
        assert result.output == ""

    def test_single(self, runner: CliRunner, shader_path_env: Path):
        result = runner.invoke(cli, ["ls"])

        assert result.exit_code == 0
        assert result.output.strip() == "shader"

    def test_multiple(
        self, runner: CliRunner, shader_path_factory: Callable[[str], Path]
    ):
        shader_path_factory("shader1")
        shader_path_factory("shader2")
        shader_path_factory("shader3")

        result = runner.invoke(cli, ["ls"])

        assert result.exit_code == 0
        line = iter(result.output.strip().splitlines())
        assert next(line).strip() == "shader1"
        assert next(line).strip() == "shader2"
        assert next(line).strip() == "shader3"
        assert next(line, None) is None

    def test_active(
        self, runner: CliRunner, shader_path_factory: Callable[[str], Path]
    ):
        shader_path_factory("shader1")
        shader2_path = shader_path_factory("shader2")
        shader_path_factory("shader3")

        hyprctl.set_screen_shader(shader2_path.as_posix())
        result = runner.invoke(cli, ["ls"])

        assert result.exit_code == 0
        line = iter(result.output.strip().splitlines())
        assert next(line).strip() == "shader1"
        assert next(line).strip() == "* shader2"
        assert next(line).strip() == "shader3"
        assert next(line, None) is None


class TestLsLong:
    def test_empty(self, runner: CliRunner):
        result = runner.invoke(cli, ["ls", "--long"])

        assert result.exit_code == 0
        assert result.output == ""

    def test_single(self, runner: CliRunner, shader_path_env: Path):
        result = runner.invoke(cli, ["ls", "--long"])

        assert result.exit_code == 0

        pattern = rf"shader +{re.escape(shader_path_env.parent.as_posix())}"
        assert re.match(pattern, result.output.strip()) is not None

    def test_multiple(
        self, runner: CliRunner, shader_path_factory: Callable[[str], Path]
    ):
        shader1_path = shader_path_factory("shader1")
        shader2_path = shader_path_factory("shader2")
        shader3_path = shader_path_factory("shader3")

        result = runner.invoke(cli, ["ls", "--long"])

        assert result.exit_code == 0
        line = iter(result.output.strip().splitlines())

        pattern = rf"shader1 +{re.escape(shader1_path.parent.as_posix())}"
        assert re.match(pattern, next(line).strip()) is not None

        pattern = rf"shader2 +{re.escape(shader2_path.parent.as_posix())}"
        assert re.match(pattern, next(line).strip()) is not None

        pattern = rf"shader3 +{re.escape(shader3_path.parent.as_posix())}"
        assert re.match(pattern, next(line).strip()) is not None

        assert next(line, None) is None

    def test_active(
        self, runner: CliRunner, shader_path_factory: Callable[[str], Path]
    ):
        shader1_path = shader_path_factory("shader1")
        shader2_path = shader_path_factory("shader2")
        shader3_path = shader_path_factory("shader3")

        hyprctl.set_screen_shader(shader2_path.as_posix())
        result = runner.invoke(cli, ["ls", "--long"])

        assert result.exit_code == 0
        line = iter(result.output.strip().splitlines())

        pattern = rf"shader1 +{re.escape(shader1_path.parent.as_posix())}"
        assert re.match(pattern, next(line).strip()) is not None

        pattern = rf"\* shader2 +{re.escape(shader2_path.parent.as_posix())}"
        assert re.match(pattern, next(line).strip()) is not None

        pattern = rf"shader3 +{re.escape(shader3_path.parent.as_posix())}"
        assert re.match(pattern, next(line).strip()) is not None

        assert next(line, None) is None



from collections.abc import Callable
from pathlib import Path

import pytest

from hyprshade.shader.core import Shader


class TestDirs:
    def test_env(self, shader_dir_env: Path, monkeypatch: pytest.MonkeyPatch):
        assert str(shader_dir_env) == Shader.dirs.env()
        monkeypatch.delenv(Shader.dirs.ENV_VAR_NAME, raising=False)
        assert Shader.dirs.env() != str(shader_dir_env)

    def test_user_hypr(self, shader_dir_user: Path):
        assert Shader.dirs.user_hypr() == str(shader_dir_user)

    def test_system(self, shader_dir_system: Path):
        assert str(shader_dir_system) == Shader.dirs.system()

    def test_all(
        self,
        shader_dir_env: Path,
        shader_dir_user: Path,
        shader_dir_system: Path,
    ):
        assert Shader.dirs.all() == list(
            map(str, [shader_dir_env, shader_dir_user, shader_dir_system])
        )


class TestConstructor:
    def test_name(self):
        assert Shader("foo")._name == "foo"

    def test_path(self, shader_path: Path):
        shader = Shader(str(shader_path))
        assert shader._name == "shader"
        assert shader._given_path == str(shader_path)
        assert shader._stale is False

    def test_nonexistent_path(self, tmp_path: Path):
        shader_path = tmp_path / "foo.glsl"
        if shader_path.exists():
            pytest.fail(f"test assumption failed: {shader_path} exists")

        shader = Shader(str(shader_path))
        assert shader._name == "foo"
        assert shader._given_path == str(shader_path)
        assert shader._stale


class TestEquality:
    def test_same_path(self, shader_path_env: Path):
        assert Shader("shader") == Shader("shader")
        assert Shader(str(shader_path_env)) == Shader(str(shader_path_env))
        assert Shader("shader") == Shader(str(shader_path_env))
        assert Shader(str(shader_path_env)) == Shader("shader")

    def test_different_path(self, shader_path_factory: Callable[[str], Path]):
        shader_path1 = shader_path_factory("foo")
        shader_path2 = shader_path_factory("bar")

        assert Shader(str(shader_path1)) == Shader("foo")
        assert Shader(str(shader_path2)) == Shader("bar")

        assert Shader("foo") != Shader("bar")
        assert Shader(str(shader_path1)) != Shader(str(shader_path2))
        assert Shader("foo") != Shader(str(shader_path2))
        assert Shader(str(shader_path1)) != Shader("bar")

    def test_other(self, shader_path_env: Path):
        assert Shader("shader") != None  # noqa: E711
        assert Shader("shader") != "shader"
        assert Shader("doesnotexist") != Shader("doesnotexist")


class TestDisplay:
    def test_str(self):
        assert str(Shader("foo")) == "foo"

    def test_repr(self):
        assert repr(Shader("foo")) == "Shader('foo')"

    def test_name(self):
        assert Shader("foo").name == "foo"

    def test_dirname(self, shader_dir_env: Path, shader_path_env: Path):
        assert Shader(str(shader_path_env)).dirname() == str(shader_dir_env.resolve())


@pytest.mark.requires_hyprland()
@pytest.mark.usefixtures("_save_screen_shader")
class TestOnOff:
    def test_on(self, shader_path_env: Path):
        Shader("shader").on()
        assert Shader.current() == Shader("shader")

    def test_on_path(self, shader_path: Path):
        Shader(str(shader_path)).on()
        assert Shader.current() == Shader(str(shader_path))

    def test_on_doesnotexist(self):
        with pytest.raises(FileNotFoundError):
            Shader("doesnotexist").on()

    def test_off(self):
        Shader.off()
        assert Shader.current() is None


class TestResolvePath:
    def test_one_path(self, shader_path_env: Path):
        assert Shader("shader")._resolve_path() == str(shader_path_env)

    def test_env_priority(
        self, monkeypatch: pytest.MonkeyPatch, shader_path: Path, tmp_path: Path
    ):
        system_path = tmp_path / "hypr/system/shaders"
        system_path.mkdir(parents=True)
        (system_path / "shader.glsl").touch()

        Shader.dirs.SYSTEM_DIR = str(tmp_path)  # type: ignore[misc]
        monkeypatch.setenv(Shader.dirs.ENV_VAR_NAME, str(shader_path.parent))

        assert Shader("shader")._resolve_path() == str(shader_path)

    def test_given_path(self, shader_path: Path):
        assert Shader(str(shader_path))._resolve_path() == str(shader_path)

    def test_notfound(self, tmp_path: Path, shader_dir_system: Path):
        with pytest.raises(FileNotFoundError):
            Shader("foo")._resolve_path()

        path_not_exist = tmp_path / "foo.glsl"
        if path_not_exist.exists():
            pytest.fail(f"test assumption failed: {path_not_exist} exists")
        with pytest.raises(FileNotFoundError):
            Shader(str(path_not_exist))._resolve_path()

    def test_ignores_cwd(self, monkeypatch: pytest.MonkeyPatch, tmp_path: Path):
        shader_path = tmp_path / "foo"
        shader_path.touch()
        monkeypatch.chdir(tmp_path)
        with pytest.raises(FileNotFoundError):
            Shader("foo")._resolve_path()



import subprocess
from pathlib import Path

import pytest

from hyprshade.shader import hyprctl

pytestmark = [
    pytest.mark.requires_hyprland(),
    pytest.mark.usefixtures("_save_screen_shader"),
]


def test_hyprctl(shader_path: Path):
    hyprctl.set_screen_shader(str(shader_path))
    assert hyprctl.get_screen_shader() == str(shader_path)


def test_clear():
    hyprctl.clear_screen_shader()
    assert hyprctl.get_screen_shader() is None


def mock_hyprctl_invalid_json(*args, **kwargs) -> subprocess.CompletedProcess:
    return subprocess.CompletedProcess(
        args=args,
        returncode=0,
        stdout='{"str": "test}',
        stderr="",
    )


def test_json_error(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr(hyprctl.subprocess, "run", mock_hyprctl_invalid_json)
    with pytest.raises(RuntimeError):
        hyprctl.get_screen_shader()



from __future__ import annotations

import click

from hyprshade.shader import Shader


@click.command(short_help="Turn off screen shader")
def off():
    """Turn off screen shader."""

    Shader.off()



from __future__ import annotations

import os
import tomllib
from os import path
from typing import cast

from more_itertools import first_true, nth, partition

from .types import ConfigDict, DefaultShadeConfig, ShaderConfig
from .utils import hypr_config_home, hyprshade_config_home


class Config:
    _dict: ConfigDict

    def __init__(self, path_: str | None = None):
        path_ = path_ or Config.get_path()
        if path_ is None:
            raise FileNotFoundError(
                "Could not find a config file; see https://github.com/loqusion/hyprshade#scheduling"
            )
        self._dict = Config._load(path_)

    @staticmethod
    def _load(path_: str) -> ConfigDict:
        with open(path_, "rb") as f:
            return cast(ConfigDict, tomllib.load(f))

    @staticmethod
    def get_path() -> str | None:
        candidates = [
            os.getenv("HYPRSHADE_CONFIG"),
            path.join(hypr_config_home(), "hyprshade.toml"),
            path.join(hyprshade_config_home(), "config.toml"),
        ]
        return first_true((c for c in candidates if c is not None), pred=path.isfile)

    def partition(self) -> tuple[list[ShaderConfig], DefaultShadeConfig | None]:
        no_default, yes_default = partition(
            lambda s: s.get("default", False), self._dict["shades"]
        )
        rest = cast(list[ShaderConfig], list(no_default))
        default = cast(DefaultShadeConfig, nth(yes_default, 0))

        assert nth(yes_default, 0) is None, "There should be only one default shade"

        return rest, default



from __future__ import annotations

import os
from typing import TYPE_CHECKING, AnyStr

if TYPE_CHECKING:
    from collections.abc import Iterator
    from datetime import time

    from _typeshed import GenericPath


def is_time_between(time_: time, start_time: time, end_time: time) -> bool:
    if end_time <= start_time:
        return start_time <= time_ or time_ <= end_time
    return start_time <= time_ <= end_time


def scandir_recursive(
    path: GenericPath[AnyStr], *, max_depth: int
) -> Iterator[os.DirEntry[AnyStr]]:
    assert max_depth >= 0

    dir_stack = []

    with os.scandir(path) as it:
        for direntry in it:
            if not direntry.is_dir():
                yield direntry
            elif max_depth > 0:
                dir_stack.append(direntry)

    while dir_stack:
        yield from scandir_recursive(dir_stack.pop(), max_depth=max_depth - 1)



from pathlib import Path

import pytest
from click.testing import CliRunner

from hyprshade.cli import cli
from hyprshade.shader import hyprctl

pytestmark = [
    pytest.mark.requires_hyprland(),
    pytest.mark.usefixtures("_save_screen_shader"),
]


def test_path(runner: CliRunner, shader_path_env: Path):
    result = runner.invoke(cli, ["on", shader_path_env.as_posix()])

    assert result.exit_code == 0
    assert hyprctl.get_screen_shader() == shader_path_env.as_posix()


def test_name(runner: CliRunner, shader_path_env: Path):
    result = runner.invoke(cli, ["on", "shader"])

    assert result.exit_code == 0
    assert hyprctl.get_screen_shader() == shader_path_env.as_posix()


def test_no_args(runner: CliRunner):
    initial_shader = hyprctl.get_screen_shader()
    result = runner.invoke(cli, ["on"])

    assert result.exit_code != 0
    assert result.exception is not None
    assert "missing argument" in result.output.lower()
    assert hyprctl.get_screen_shader() == initial_shader


def test_invalid_shader(
    runner: CliRunner,
    shader_dir_env: Path,
    shader_dir_user: Path,
    shader_dir_system: Path,
):
    initial_shader = hyprctl.get_screen_shader()
    result = runner.invoke(cli, ["on", "invalid"])

    assert result.exit_code != 0
    assert isinstance(result.exception, FileNotFoundError)
    assert hyprctl.get_screen_shader() == initial_shader



import re
from pathlib import Path

import pytest

from hyprshade.config import utils


class TestXDG:
    def test_uses_xdg_config_home(
        self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch
    ):
        monkeypatch.setenv("XDG_CONFIG_HOME", str(tmp_path))
        assert utils.xdg_config_home() == str(tmp_path)

    def test_uses_home(self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.delenv("XDG_CONFIG_HOME", raising=False)
        monkeypatch.setenv("HOME", str(tmp_path))
        assert utils.xdg_config_home() == str(tmp_path / ".config")

    def test_raises_when_cannot_be_determined(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.delenv("XDG_CONFIG_HOME", raising=False)
        monkeypatch.delenv("HOME", raising=False)
        with pytest.raises(
            ValueError, match=re.escape("$HOME environment variable is not set")
        ):
            utils.xdg_config_home()


class TestConfigHomes:
    def test_hypr(self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setenv("XDG_CONFIG_HOME", str(tmp_path))
        assert utils.hypr_config_home() == str(tmp_path / "hypr")

    def test_hyprshade(self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setenv("XDG_CONFIG_HOME", str(tmp_path))
        assert utils.hyprshade_config_home() == str(tmp_path / "hyprshade")

    def test_systemd_user(self, tmp_path: Path, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setenv("XDG_CONFIG_HOME", str(tmp_path))
        assert utils.systemd_user_config_home() == str(tmp_path / "systemd/user")



from __future__ import annotations

from bisect import bisect
from typing import final

import click

from hyprshade.shader import Shader

from .utils import ls_dirs


@final
class ShaderWithMeta(Shader):
    _is_current: bool
    _is_in_shader_paths: bool | None

    def __init__(self, shader_name_or_path: str):
        super().__init__(shader_name_or_path)
        self._is_current = False
        self._is_in_shader_paths = None

    @property
    def is_current(self):
        return self._is_current

    @property
    def is_in_shader_paths(self):
        is_in_shader_paths = self._is_in_shader_paths
        if is_in_shader_paths is None:
            raise RuntimeError("ShaderWithMeta.is_in_shader_paths is not set")
        return is_in_shader_paths

    @classmethod
    def _current(cls) -> ShaderWithMeta | None:
        shader = super().current()
        if shader is None:
            return None
        shader = cls(shader._resolve_path())
        shader._is_current = True
        return shader

    @staticmethod
    def get_shaders_list() -> list[ShaderWithMeta]:
        current = ShaderWithMeta._current()
        shaders = list(map(ShaderWithMeta, ls_dirs(Shader.dirs.all())))
        if current:
            i = bisect(shaders, current.name, key=lambda s: s.name)
            is_current_in_shaders = i != 0 and shaders[i - 1] == current
            if is_current_in_shaders:
                shaders[i - 1]._is_current = True
                shaders[i - 1]._is_in_shader_paths = True
            else:
                current._is_in_shader_paths = False
                shaders.insert(i, current)
        return shaders


@click.command(short_help="List available screen shaders")
@click.option("-l", "--long", is_flag=True, help="Long listing format")
def ls(long: bool):
    """List available screen shaders."""

    shaders = ShaderWithMeta.get_shaders_list()
    if not shaders:
        return
    width = max(map(len, map(str, shaders))) + 1

    for shader in shaders:
        c = "*" if shader.is_current else " "
        if long:
            click.echo(f"{c} {shader!s:{width}} {shader.dirname()}")
            continue
        if shader.is_current and not shader.is_in_shader_paths:
            click.echo(f"{c} {shader!s}  ({shader.dirname()})")
            continue
        click.echo(f"{c} {shader!s}")



from __future__ import annotations

from typing import TYPE_CHECKING

import click

from .utils import convert_to_shader

if TYPE_CHECKING:
    from hyprshade.shader import Shader


@click.command(short_help="Turn on screen shader")
@click.argument("shader", callback=convert_to_shader)
def on(shader: Shader):
    """Turn on screen shader."""

    shader.on()



from __future__ import annotations

from datetime import datetime

import click

from hyprshade.config.schedule import Schedule
from hyprshade.shader import Shader


@click.command(short_help="Set screen shader on schedule")
@click.pass_context
def auto(ctx: click.Context):
    """Set screen shader based on schedule.

    Requires a schedule to be specified in hyprshade.toml.
    """

    t = datetime.now().time()
    shader = Schedule.from_config().scheduled_shader(t)

    if shader:
        shader.on()
    else:
        Shader.off()



from __future__ import annotations

from datetime import time
from typing import Literal, NotRequired, TypedDict

TimeRange = tuple[time, time]


class ShaderConfig(TypedDict):
    name: str
    start_time: time
    end_time: NotRequired[time]


class DefaultShadeConfig(TypedDict):
    name: str
    default: Literal[True]


class ConfigDict(TypedDict):
    shades: list[ShaderConfig | DefaultShadeConfig]



from __future__ import annotations

import json
import subprocess
import textwrap
from json import JSONDecodeError
from typing import Final

import click

EMPTY_STR: Final = "[[EMPTY]]"


class HyprctlError(Exception):
    def __init__(self, error: subprocess.CalledProcessError, *args, **kwargs):
        command = " ".join(error.cmd)
        stdout = str(error.stdout).strip() or "<empty>"
        stderr = str(error.stderr).strip() or "<empty>"
        message = f"""hyprctl returned a non-zero exit code.

{click.style("command", fg="red")}:
{textwrap.indent(command, " " * 4)}

{click.style("stdout", fg="red")}:
{textwrap.indent(stdout, " " * 4)}

{click.style("stderr", fg="red")}:
{textwrap.indent(stderr, " " * 4)}"""

        super().__init__(message, *args, **kwargs)


def set_screen_shader(shader_path: str) -> None:
    try:
        subprocess.run(
            ["hyprctl", "keyword", "decoration:screen_shader", shader_path],
            capture_output=True,
            check=True,
            encoding="utf-8",
        )
    except subprocess.CalledProcessError as e:
        raise HyprctlError(e) from e


def clear_screen_shader() -> None:
    set_screen_shader(EMPTY_STR)


def get_screen_shader() -> str | None:
    try:
        hyprctl_pipe = subprocess.run(
            ["hyprctl", "-j", "getoption", "decoration:screen_shader"],
            capture_output=True,
            check=True,
            encoding="utf-8",
        )
        shader_json = json.loads(hyprctl_pipe.stdout)
    except subprocess.CalledProcessError as e:
        raise HyprctlError(e) from e
    except JSONDecodeError as e:
        stdout = str(hyprctl_pipe.stdout).strip() or "<empty>"
        stderr = str(hyprctl_pipe.stderr).strip() or "<empty>"
        message = f"""hyprctl returned invalid JSON.
This is likely a bug in Hyprland; go bug Vaxry about it (nicely :)).

{click.style("stdout", fg="red")}:
{textwrap.indent(stdout, " " * 4)}

{click.style("stderr", fg="red")}:
{textwrap.indent(stderr, " " * 4)}"""

        raise RuntimeError(message) from e

    shader = str(shader_json["str"]).strip()
    if shader == EMPTY_STR:
        return None

    return shader



from __future__ import annotations

import logging
from typing import Final

import click

from .auto import auto
from .current import current
from .install import install
from .ls import ls
from .off import off
from .on import on
from .toggle import toggle

COMMANDS: Final = [
    auto,
    current,
    install,
    ls,
    off,
    on,
    toggle,
]


@click.group()
@click.version_option()
@click.option("-v", "--verbose", is_flag=True, help="Enable verbose output")
def cli(verbose: bool):
    level = logging.DEBUG if verbose else logging.WARNING
    logging.basicConfig(level=level)


for command in COMMANDS:
    cli.add_command(command)


def main():
    try:
        return cli()
    except Exception as e:
        if logging.getLogger().getEffectiveLevel() <= logging.DEBUG:
            raise e
        click.echo(f"{click.style('Error', fg='red')}: {e!s}", err=True)
        return 1



from __future__ import annotations

import os
from os import path


def xdg_config_home():
    config_home = os.getenv("XDG_CONFIG_HOME")
    if config_home is None:
        home = os.getenv("HOME")
        if home is None:
            raise ValueError("$HOME environment variable is not set")
        config_home = path.join(home, ".config")
    return config_home


def hypr_config_home():
    return path.join(xdg_config_home(), "hypr")


def hyprshade_config_home():
    return path.join(xdg_config_home(), "hyprshade")


def systemd_user_config_home():
    return path.join(xdg_config_home(), "systemd/user")
